/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * Copyright (C) 2006-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://openbluesky.kenai.com
 * SCM: https://bitbucket.org/tidalwave/openbluesky-src
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.aspect;

import javax.annotation.Nonnull;
import java.util.HashSet;
import java.util.Set;
import org.openide.util.Lookup;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.util.Task;
import it.tidalwave.aspect.Aspect;
import it.tidalwave.aspect.MultiAspect;
import it.tidalwave.aspect.AspectFactory;

/***********************************************************************************************************************
 *
 * Since in a NetBeans environment it's likely that {@code Aspect}s are provided dynamically by means of a
 * {@link Lookup}, this class provides facility methods for writing less code:
 *
 * <pre>
 * Lookup lookup = ... // the Lookup which contains the AspectFactories
 * Aspects.run(lookup, new Aspect.Task&lt;String, NotFoundException&gt;()
 *   {
 *     &#64;Override
 *     public String run()
 *       throws NotFoundException
 *       {
 *         // do something
 *         return "something";
 *       }
 *   });
 * </pre>
 * 
 * @author Fabrizio Giudici
 * @version $Id$
 * @experimental
 *
 **********************************************************************************************************************/
public final class Aspects
  {
    private static final String CLASS = Aspects.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    private Aspects()
      {
      }

    /*******************************************************************************************************************
     *
     * Searches for {@link AspectFactory} instances in the {@link Lookup} provided by the given provider, creates their
     * product {@link Aspect}s and runs the specified task with them.
     *
     * @param  lookupProvider   the {@code Lookup} provider
     * @param  task             the {@code Task} to run
     *
     ******************************************************************************************************************/
    @Nonnull
    public static <T, E extends Throwable> T run (final @Nonnull Lookup.Provider lookupProvider,
                                                  final @Nonnull Task<T, E> task)
      throws E
      {
        return run(lookupProvider.getLookup(), task);
      }

    /*******************************************************************************************************************
     *
     * Searches for {@link AspectFactory} instances in the provided {@link Lookup}, creates their
     * product {@link Aspect}s and runs the specified task with them.
     *
     * @param  lookupProvider   the {@code Lookup} provider
     * @param  task             the {@code Task} to run
     *
     ******************************************************************************************************************/
    @Nonnull
    public static <T, E extends Throwable> T run (final @Nonnull Lookup lookup,
                                                  final @Nonnull Task<T, E> task)
      throws E
      {
        final Set<Aspect> aspects = new HashSet<Aspect>();
        
        for (final AspectFactory aspectFactory : lookup.lookupAll(AspectFactory.class))
          {
            aspects.add(aspectFactory.createAspect());
          }

        return new MultiAspect(aspects).run(task);
      }
  }
