/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * Copyright (C) 2006-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://openbluesky.kenai.com
 * SCM: https://bitbucket.org/tidalwave/openbluesky-src
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.util;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import org.openide.util.Lookup;
import org.openide.util.lookup.ServiceProvider;
import it.tidalwave.util.Id;
import it.tidalwave.role.IdFactory;

/***********************************************************************************************************************
 *
 * A default implementation of {@link IdFactory} which implements a pluggable mechanism in which a different factory
 * can be provided for each class. This implementation searches for instances of {@link IdFactoryDelegate}s installed
 * in the system by means of the {@code META-INF/services} facility and uses the one which declares to support the
 * required class. A default fallback case is provided, which creates UUID-based ids.
 *
 * @author  Fabrizio Giudici
 * @version $Id: $
 * @stable
 *
 **********************************************************************************************************************/
@ServiceProvider(service=IdFactory.class)
public final class DefaultIdFactory implements IdFactory
  {
    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Id createId()
      {
        return createId(Object.class);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Id createId (final @Nonnull Class<?> clazz)
      {
        return internalCreateId(clazz, null);
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Id createId (final @Nonnull Class<?> clazz, final @Nonnull Object object)
      {
        return internalCreateId(clazz, object);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private Id internalCreateId (final @Nonnull Class<?> clazz, final @CheckForNull Object object)
      {
        // FIXME: cache for performance
        // FIXME: the list should be ordered from subclasses to superclasses for managing inheritance
        for (final IdFactoryDelegate<?> idFactoryDelegate : Lookup.getDefault().lookupAll(IdFactoryDelegate.class))
          {
            if (idFactoryDelegate.getManagedClass().isAssignableFrom(clazz))
              {
                return idFactoryDelegate.createId(object);
              }
          }

        throw new RuntimeException("There's no IdFactoryDelegate for Object installed in the system");
      }
  }
