/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * Copyright (C) 2006-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://openbluesky.kenai.com
 * SCM: https://bitbucket.org/tidalwave/openbluesky-src
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.util;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import java.util.concurrent.atomic.AtomicBoolean;
import java.awt.image.BufferedImage;
import javax.swing.Icon;
import org.openide.util.ImageUtilities;
import org.openide.util.RequestProcessor;
import it.tidalwave.role.IconProvider;
import it.tidalwave.role.MutableIconProvider;
import it.tidalwave.role.MutableIconProviderSupport;
import it.tidalwave.util.NotFoundException;

/***********************************************************************************************************************
 *
 * An implementation of {@link MutableIconProvider} which creates the icon on demand and in a background thread. The icon
 * must be created by subclasses in the method {@link #createIcon(int)}. The
 * method can throw an exception to mean that the icon can't be created, in which case a default icon will be provided.
 * It is possible to specify another {@link IconProvider} for the default icon.
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 * @stable
 *
 **********************************************************************************************************************/
public abstract class DeferredIconProvider extends MutableIconProviderSupport
  {
    private static final Icon EMPTY_ICON = ImageUtilities.image2Icon(new BufferedImage(16, 16, BufferedImage.TYPE_4BYTE_ABGR));

    @CheckForNull
    private Icon icon;

    @CheckForNull
    private final IconProvider defaultIconProvider;

    private final AtomicBoolean loaded = new AtomicBoolean(false);

    /*******************************************************************************************************************
     *
     * Creates an instance which uses an empty icon in case of default.
     *
     ******************************************************************************************************************/
    public DeferredIconProvider()
      {
        this.defaultIconProvider = null;
      }

    /*******************************************************************************************************************
     *
     * Creates an instance with uses a specific provider of default icons.
     *
     * @param  defaultIconProvider  the default icon provider
     *
     ******************************************************************************************************************/
    public DeferredIconProvider (final @Nonnull IconProvider defaultIconProvider)
      {
        this.defaultIconProvider = defaultIconProvider;
      }

    /*******************************************************************************************************************
     *
     * {@inheritDoc}
     *
     ******************************************************************************************************************/
    @Override @Nonnull
    public Icon getIcon (final @Nonnegative int size)
      {
        if (!loaded.getAndSet(true))
          {
            RequestProcessor.getDefault().post(new Runnable()
              {
                @Override
                public void run()
                  {
                    final Icon oldIcon = icon;

                    try
                      {
                        icon = createIcon(size);
                      }
                    catch (NotFoundException e)
                      {
                        icon = getDefaultIcon(size);
                      }

                    if (oldIcon != icon)
                      {
                        fireIconChange(oldIcon, icon); // FIXME: should be in the EDT?
                      }
                  }
              });
          }

        return (icon != null) ? icon : getDefaultIcon(size);
      }

    /*******************************************************************************************************************
     *
     * Creates the icon. This method must be implemented by concrete subclasses. It can throw a
     * {@link NotFoundException} if the icon can't be created, in which case a default icon will be used.
     * 
     * @param   type  the icon type
     * @param   size  the icon size
     * @return        the icon
     * @throws        NotFoundException  if the icon can't be created
     *
     ******************************************************************************************************************/
    @Nonnull
    protected abstract Icon createIcon (final @Nonnegative int size)
      throws NotFoundException;

    /*******************************************************************************************************************
     *
     * Creates a default icon, calling the delegate if provided, otherwise returning an empty icon.
     * 
     * @param   type  the icon type
     * @param   size  the icon size
     * @return        the icon
     *
     ******************************************************************************************************************/
    @Nonnull
    private Icon getDefaultIcon (final @Nonnegative int size)
      {
        return (defaultIconProvider != null) ? defaultIconProvider.getIcon(size) : EMPTY_ICON;
      }
  }
