/***********************************************************************************************************************
 *
 * OpenBlueSky - NetBeans Platform Enhancements
 * Copyright (C) 2006-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://openbluesky.kenai.com
 * SCM: https://bitbucket.org/tidalwave/openbluesky-src
 *
 **********************************************************************************************************************/
package it.tidalwave.netbeans.util;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.concurrent.ThreadSafe;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import javax.inject.Provider;
import java.util.concurrent.atomic.AtomicReference;
import java.awt.EventQueue;
import org.openide.util.Lookup;

/***********************************************************************************************************************
 *
 * This utility class provides static factory methods for retrieving services registered in the global {@link Lookup}
 * or {@link TopComponent}s. For instance, if {@code MyService} is a registered service, the following code is
 * roughly equivalent to {@code MyService service = Lookup.getDefault().lookup(MyService.class)}:
 *
 * <pre>
 * MyService service = Locator.find(MyService.class);
 * </pre>
 *
 * In case of missing services, while {@code Lookup} just returns a {@code null}, {@code Locator} throws a
 * {@link Locator.NotFoundException}.
 *
 * In case of {@code TopComponent}s, the following code is roughly equivalent to
 * {@code MyTopComponent topComponent = MyTopComponent.findInstance()}:
 *
 * <pre>
 * MyTopComponent topComponent = Locator.find(TopComponent.class);
 * </pre>
 *
 * In case a {@code TopComponent} implements an interface (e.g. {@code MyInterface}), it is possible to use the
 * interface as the argument to {@code Locator.find()}:
 *
 * <pre>
 * MyInterface myStuff = Locator.find(MyInterface.class);
 * </pre>
 *
 * if the special {@code Locator.ImplementedBy} annotation is used:
 *
 * <pre>
 * import it.tidalwave.netbeans.util.Locator.ImplementedBy;
 *
 * &#64;ImplementedBy(MyTopComponent.class)
 * public interface MyInterface
 *   {
 *     ...
 *   }
 *
 * public MyTopComponent extends TopComponent implements MyInterface
 *   {
 *     ...
 *   }
 * </pre>
 *
 * It is possible to have {@code Locator} to return a {@link Provider} of required service instead of the service
 * itself:
 *
 * <pre>
 * import javax.inject.Provider;
 *
 * Provider&lt;MyService&gt; serviceProvider = Locator.createProviderFor(MyService.class);
 * Provider&lt;MyTopComponent&gt; topComponentProvider = Locator.createProviderFor(MyTopComponent.class);
 *
 * ...
 *
 * MyService service = serviceProvider.get();
 * MyTopComponent topComponent = topComponentProvider.get();
 * </pre>
 * 
 * @author  Fabrizio Giudici
 * @version $Id$
 * @stable
 *
 **********************************************************************************************************************/
@ThreadSafe
public final class Locator<T>
  {
    /*******************************************************************************************************************
     *
     * This annotation makes it possible to search for {@link TopComponent}s that implements a given interface. Please
     * see the documentation of {@link Locator} for more information.
     *
     ******************************************************************************************************************/
    @Retention(RetentionPolicy.RUNTIME)
    public @interface ImplementedBy
      {
        Class<?> value();
      }

    /*******************************************************************************************************************
     *
     * This exception is thrown by a {@link Locator} (or a {@link Provider} created by a {@code Locator} when the
     * required service is not found.
     *
     ******************************************************************************************************************/
    public static class NotFoundException extends RuntimeException
      {
        public NotFoundException (final @Nonnull Class<?> serviceClass)
          {
            super("Not found: " + serviceClass);
          }

        public NotFoundException (final @Nonnull String name)
          {
            super("Not found: " + name);
          }
      }

    // Replacement for TopComponent.class, since we don't want a static dependency on org.openide.windows.
    @CheckForNull
    private final static Class<?> TOP_COMPONENT_CLASS;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    static
      {
        Class<?> clazz = null;

        try
          {
            final ClassLoader classLoader = Lookup.getDefault().lookup(ClassLoader.class);
            clazz = classLoader.loadClass("org.openide.windows.TopComponent");
          }
        catch (ClassNotFoundException ex)
          {
            clazz = null;
          }

        TOP_COMPONENT_CLASS = clazz;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private Locator()
      {
      }
    
    /*******************************************************************************************************************
     *
     * Finds a service registered in the global {@link Lookup} or a {@link TopComponent}. Please look at the class
     * javadoc for more information.
     *
     * @param  serviceClass   the service class
     * @return                the service
     * @throws                NotFoundException if the service can't be found
     *
     ******************************************************************************************************************/
    @Nonnull
    public static <T> Provider<T> createProviderFor (final @Nonnull Class<T> serviceClass)
      {
        return new Provider<T>()
          {
            @Override @Nonnull
            public T get()
              {
                return find(serviceClass);
              }
          };
      }

    /*******************************************************************************************************************
     *
     * @deprecated Use {@link #createProviderFor(Class)} instead.
     *
     ******************************************************************************************************************/
    @Nonnull @Deprecated
    public static <T> Provider<T> findProvider (final @Nonnull Class<T> serviceClass)
      {
        return createProviderFor(serviceClass);
      }

    /*******************************************************************************************************************
     *
     * Creates a {@link Provider} that can be used to find a service registered in the global {@link Lookup} or a
     * {@link TopComponent}. Please look at the class javadoc for more information.
     *
     * @param  serviceClass   the service class
     * @return                the {@code Provider}
     *
     ******************************************************************************************************************/
    @Nonnull
    public static <T> T find (final @Nonnull Class<T> serviceClass)
      {
        Class<? extends T> serviceClass2 = serviceClass;
        final ImplementedBy implementedBy = serviceClass.getAnnotation(ImplementedBy.class);

        if (implementedBy != null)
          {
            serviceClass2 = (Class<? extends T>)implementedBy.value();
          }

        if ((TOP_COMPONENT_CLASS != null) && TOP_COMPONENT_CLASS.isAssignableFrom(serviceClass2))
          {
            return topComponentFind(serviceClass2);
          }
        else
          {
            return simpleFind(serviceClass2);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static <T> T simpleFind (final @Nonnull Class<T> serviceClass)
      {
        final T service = Lookup.getDefault().lookup(serviceClass);

        if (service == null)
          {
            throw new NotFoundException(serviceClass);
          }

        return service;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static synchronized <T> T topComponentFind (final @Nonnull Class<T> topComponentClass)
      {
        final AtomicReference<Object> topComponent = new AtomicReference<Object>();

        final Runnable runnable = new Runnable()
          {
            @Override
            public void run()
              {
                try
                  {
                    final Method method = topComponentClass.getMethod("findInstance");
                    topComponent.set(method.invoke(null));
                  }
                catch (RuntimeException e)
                  {
                    throw new RuntimeException("Cannot lookup " + topComponentClass, e);
                  }
                catch (Exception e)
                  {
                    throw new RuntimeException("Cannot lookup " + topComponentClass, e);
                  }
              }
          };

        if (EventQueue.isDispatchThread())
          {
            runnable.run();
          }
        else
          {
            try
              {
                EventQueue.invokeAndWait(runnable);
              }
            catch (InterruptedException e)
              {
                throw new RuntimeException(e);
              }
            catch (InvocationTargetException e)
              {
                throw new RuntimeException(e);
              }
          }

        if (topComponent.get() == null)
          {
            throw new NotFoundException(topComponentClass.getSimpleName());
          }

        return (T)topComponent.get();
      }
  }
